from __future__ import print_function
import os
import sys
import stat
import shutil
import shlex
import logging
import subprocess
import threading
from threading import Timer
from distutils.version import LooseVersion
from mcutk.exceptions import ProcessTimeout

PY = sys.version_info[0]

"""
This module provide some useful functions.
"""

def to_hex(val):
    """ Safe to convert to hex"""

    if type(val) == str:
        return val
    return format(val, 'x')


def _readerthread(fh, buffer):
    while True:
        stdoutdata = fh.readline()
        if stdoutdata:
            sys.stdout.write(stdoutdata)
            sys.stdout.flush()
            buffer.append(stdoutdata)
        else:
            break


def run_command(cmd, cwd=None, shell=True, stdout=False, timeout=None, need_raise=False):
    """Run a command with a timeout timer and capture it's console output.

    This method wrapped subprocess.Popen, when you need to capture console output, just make
    sure **stdout** is True.

    Arguments:
        cmd -- {str or list} command string or list, like subprocess
        cwd -- {str} process work directory.
        shell -- {boolean} Use shell or not.
        stdout -- {boolean} Capture stdout and print it in real time, default: False.
        timeout -- {int} timeout in seconds, default: None
        need_raise -- {boolean} a switch to disable raising ProcessTimeout exception,
                    just logging it as an error message, default: False.

    Returns:
        Tuple -- (returncode, output)
    """
    output = ""
    returncode, timer, error_message = None, None, None
    timer_result = {"is_timeout": False}

    if shell:
        # python documentation:
        # On Windows, The shell argument (which defaults to False) specifies
        # whether to use the shell as the program to execute.
        # If shell is True, it is recommended to pass args as a string rather
        # than as a sequence.
        if isinstance(cmd, list):
            cmd = " ".join(cmd)
    else:
        # Windows platform, convert cmd to list to will lead out the slash issue.
        if os.name == 'nt':
            if not isinstance(cmd, list):
                cmd = shlex.split(cmd)

    kwargs = {
        "stdout": subprocess.PIPE if stdout else None,
        "stderr": subprocess.STDOUT,
        "cwd": cwd,
        "shell": shell
    }

    if PY > 2:
        kwargs['encoding'] = 'utf8'

    try:
        process = subprocess.Popen(cmd, **kwargs)

        # start timer
        if timeout:
            timer = Timer(timeout, _timeout_trigger, args=(process, timer_result, ))
            timer.start()

        if stdout:
            output = []
            stdout_thread = threading.Thread(
                target=_readerthread,
                args=(process.stdout, output))
            stdout_thread.setDaemon(True)
            stdout_thread.start()

            process.wait()
            stdout_thread.join()
            output = ''.join(output)
        else:
            output, error = process.communicate()

        returncode = process.returncode

        if returncode != 0:
            error_message = 'Error: {0}\n  exit code:  {1}\n'.format(cmd, process.pid)
            if output:
                error_message += ' console output: %s'%(output)
            logging.debug(error_message)

    except OSError as emsg:
        logging.exception(emsg)

    finally:
        if timer:
            timer.cancel()

        if timer_result['is_timeout']:
            msg = "process(pid %s) timeout(%ss)" % (process.pid, timeout)
            if need_raise:
                raise ProcessTimeout(msg)
            else:
                logging.error(msg)

    return returncode, output



def _timeout_trigger(pro, result):
    """Timeout will kill the group processes.

    [Timeout will kill the group processes]

    Arguments:
        pro {Popen object} -- process
    """

    if os.name == "nt":
        subprocess.Popen("TASKKILL /F /PID {pid} /T".format(pid=pro.pid))
    else:
        pro.kill()

    if isinstance(result, dict):
        result['is_timeout'] = True


def rmtree(path):
    """Remove directory tree. If failed , it will check the access and force
    to close unclosed handler, then try remove agagin.
    """
    try:
        shutil.rmtree(path)
    except Exception:
        # Is the error an access error ?
        if not os.access(path, os.W_OK):
            os.chmod(path, stat.S_IWUSR)

        # Readonly on windows
        if os.name == "nt":
            subprocess.check_call(('attrib -R ' + path + '\\* /S').split())

        shutil.rmtree(path)



def onerrorHandler(func, path, exc_info):
    """Error handler for ``shutil.rmtree``.

    If the error is due to an access error (read only file)
    it attempts to add write permission and then retries.

    If the error is for another reason it re-raises the error.

    Usage : ``shutil.rmtree(path, onerror=onerror)``
    """
    if not os.access(path, os.W_OK):
        # Is the error an access error ?
        os.chmod(path, stat.S_IWUSR)
        func(path)


def copydir(root_src_dir, root_dst_dir):
    """Copy directory to dst dir."""
    for src_dir, dirs, files in os.walk(root_src_dir):
        dst_dir = os.path.normpath(src_dir.replace(root_src_dir, root_dst_dir, 1))
        print("copying %s"%dst_dir)
        if not os.path.exists(dst_dir):
            os.makedirs(dst_dir)
        for file_ in files:
            src_file = os.path.normpath(os.path.join(src_dir, file_))
            dst_file = os.path.normpath(os.path.join(dst_dir, file_))
            if os.path.exists(dst_file):
                os.chmod(dst_file, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO) # 0777
                os.remove(dst_file)
            shutil.copy(src_file, dst_file)


def get_max_version(version_pool):
    versions = [(ver[0], LooseVersion(str(ver[1]))) for ver in version_pool]
    versions.sort(key=lambda x: x[1])
    return versions[-1]



from __future__ import print_function
import os
import io
import glob
import logging
import subprocess

from distutils.version import LooseVersion
try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse
try:
    import ConfigParser as configparser
except ImportError:
    import configparser

from mcutk import util
from mcutk.exceptions import CodeSyncError, BranchError, ProcessTimeout


bin_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'bin')


def _get_repo_url(path):
    ret, output = util.run_command('git config --get remote.origin.url', cwd=path, stdout=True, shell=True)
    if ret != 0:
        logging.warning('Not a valid git repository.')
        return None

    return output.replace('\\n', '').strip()


class Repo(object):
    """Class Repo provide an abstract model access for git repositories, it is warpper
    based on git tool. Many useful built in abstactions or shortcut functions make the
    git call is very easy so that you don't need to handle process call by subprocess
    or others.

    Exmaple 1:
        >>> repo = Repo("ssh://exmaple.git")
        >>> repo.is_ready
        >>> repo.Clone("/home/")

    Exmaple 2:
        >>> repo = Repo.frompath("C:/mcu-sdk-2.0")
        >>> print repo.repo_url
        >>> print repo.is_ready
        >>> print repo.get_branch_name()
        >>> print repo.get_latest_commit_message()
        >>> print repo.get_submodules()
        >>> repo.checkout_branch("dev")
        >>> repo.exec_commands(["git status", "git --version"])
    """

    @classmethod
    def frompath(cls, path):
        """Return a repo object from specific path."""

        url = _get_repo_url(path)
        if url:
            return cls(url, path)

    @classmethod
    def fromurl(cls, url, parent_dir):
        """Return a repo object with url and parent_dir."""

        repo = cls(url)
        repo.repo_dir = os.path.join(parent_dir, repo.repo_name)
        return repo

    def __init__(self, repo_url, repo_dir=None):
        """Create a repo instance.

        params:
            - repo_url: repo remote address.
            - repo_dir: the directory of repo on your local file system.
        """
        self.repo_dir = repo_dir
        self.repo_url = repo_url
        self.occupations = dict()
        self.repo_name = repo_url.split("/")[-1].replace(".git", "")
        self.repo_domain = repo_url.replace(self.repo_name+".git", "")

    def __str__(self):
        return "Repo(name=%s, dir=%s)"%(self.repo_name, self.repo_dir)

    def __exec(self, cmd, cwd=None, t=3600, retry=2):
        if cwd is None:
            cwd = self.repo_dir
        error = None
        for _ in range(retry):
            error = None
            logging.info(cmd)
            try:
                noerror, output = util.run_command(cmd, cwd=cwd, stdout=True, timeout=t)
                if noerror == 0:
                    break
            except ProcessTimeout as e:
                error = e
            logging.info("retry again..")
        else:
            raise CodeSyncError("Run command {0} failure".format(cmd))

        if error:
            raise
        return output

    @property
    def is_ready(self):
        if not os.path.exists(str(self.repo_dir)):
            return False

        url_c = _get_repo_url(self.repo_dir)

        if not url_c:
            return False

        return True

    def check_url(self):
        """Check and compare url whether is match."""

        url_c = _get_repo_url(self.repo_dir)

        if not url_c:
            return False

        if urlparse(self.repo_url).path == urlparse(url_c).path:
            return True

        else:
            logging.warning("URL is mismatch, local url is: %s, but object is: %s",
                            url_c, self.repo_url)
            return False

    def exec_cmd(self, cmd, cwd=None, timeout=3600, retry=2):
        """Execute a command."""

        return self.__exec(cmd, cwd, timeout, retry)

    def exec_cmds(self, cmds):
        """Execute a list of commands."""

        for cmd in cmds:
            self.__exec(cmd)

    def get_latest_commit_message(self, repo_dir=None):
        """Return the commit message on the head

        Default work directory is the main repo.
        """
        if not repo_dir:
            repo_dir = self.repo_dir

        return self.__exec("git log -n 1", cwd=repo_dir, t=1)

    def get_branch_name(self, repo_dir=None):
        """Return branch name.

        Default work directory is the main repo.
        """
        if not repo_dir:
            repo_dir = self.repo_dir
        return self.__exec("git rev-parse --abbrev-ref HEAD", cwd=repo_dir, t=1).replace("\n", "").strip()

    def get_submodules(self, repo_dir=None):
        """Return a dictionary about all submodules.

        Default return the main submodules.
        """
        if not repo_dir:
            repo_dir = self.repo_dir

        gitmodules = os.path.join(repo_dir, ".gitmodules")
        if os.path.exists(gitmodules) is False:
            return None

        mod_parser = configparser.RawConfigParser()
        with open(gitmodules) as f:
            content = f.read().replace("\t", "").decode('utf-8')
            fakefile = io.StringIO(content)

        mod_parser.readfp(fakefile, filename=gitmodules)
        modules = mod_parser.sections()
        modules_dict = dict()

        for modname in modules:
            url = mod_parser.get(modname, "url")
            path = os.path.join(self.repo_dir, mod_parser.get(modname, "path"))
            modules_dict[modname.replace("submodule ", "").replace("\"", "")] = Repo(url, path)

        return modules_dict

    def update_submodules(self):
        """Update submodules
            functions instruction:
                # update submodules
                # 1> get commit message
                # 2> get current branch
                # 3> if need, update gitsubmodules configurations
                # 4> git submodule deinit -f .
                # 5> git submodule update --init
        """
        version_string = self.exec_cmd('git --version')
        git_version = version_string.split('.windows')[0].replace("git version", '').strip()

        # git version < 2.9.0  not support --jobs
        if LooseVersion(git_version) >= LooseVersion("2.9.0"):
            cmd = 'git submodule update --init --force --recursive --jobs 4'
        else:
            cmd = 'git submodule update --init --force --recursive'

        try:
            self.exec_cmd(cmd)
        except CodeSyncError:
            Repo.remove_index_lock(self.repo_dir)
            self.exec_cmd('git submodule deinit -f .')
            self.exec_cmd(cmd)


    def get_head_hash(self):
        """Return commit hash from head."""
        return self.__exec("git rev-parse HEAD", t=1).replace("\n", "").strip()

    def get_history_abbrev(self, since="1.weeks"):
        """Useful options for git log --pretty=format
            Option	Description of Output
            %H Commit hash
            %h Abbreviated commit hash
            %T Tree hash
            %t Abbreviated tree hash
            %an Author name
            %ae Author e-mail
            %ad Author date (format respects the -date= option)
            %ar Author date, relative
            %cn Committer name
            %ce Committer email
            %cd Committer date
            %cr Committer date, relative
            %s Subject

        Reference:
            https://git-scm.com/book/tr/v2/Git-Basics-Viewing-the-Commit-History
            https://git-scm.com/docs/git-log
        """
        if since.split(".")[1] not in ["days", "weeks"]:
            raise ValueError("argument is not valid format,")

        return util.run_command('git log --pretty=format:"#### %h - %cr - %an: %s" --since={0} --stat'.format(since),
                                cwd=self.repo_dir)[1]

    def get_history(self, since="1.weeks"):
        """Extract commits history details.

            git log -p --since=1.weeks
        """
        if since.split(".")[1] not in ["days", "weeks"]:
            raise ValueError("argument is not valid format,")

        return util.run_command('git log -p --since={0}'.format(since),
                                cwd=self.repo_dir)[1]

    def get_remote_branches(self):
        """Fetch all of branches from remote.

        Returns: A list of branches.
        """
        content = util.run_command("git ls-remote --heads", cwd=self.repo_dir, stdout=True, timeout=300)[1]
        branches = [line.split("heads/")[1].strip() for line in content.split("\n") \
                    if "refs/heads/" in line]

        return branches

    def clone(self, clone_path='.', branch_name=None, foldername=None):
        """Clone specific branch to specific directory.

        Arguments:
            clone_path {str} -- directory to clone

        Keyword Arguments:
            branch_name {str} -- branch name (default: {'master'})
            foldername {str} -- folder name (default: {None})

        Returns:
            repo_dir

        Documentation about git clone:
            --depth=<depth>
                Create a shallow clone with a history truncated to the specified number of commits.
                Implies --single-branch unless --no-single-branch is given to fetch the histories
                near the tips of allbranches. If you want to clone submodules shallowly also pass
                --shallow-submodules.


            --branch <name>
            -b <name>
                Instead of pointing the newly created HEAD to the branch pointed to by the cloned
                repository's HEAD, point to <name> branch instead. In a non-bare repository, this
                is the branch that will be checked out. --branch can also take tags and detaches
                the HEAD at that commit in the resulting repository.
        """
        logging.info("prepare to clone repo")

        if self.repo_dir and clone_path == '.':
            clone_path = os.path.dirname(self.repo_dir)

        # clone single branch, instead of all branches
        if branch_name:
            clone_cmd = 'git clone --depth 10 -b {0} {1}'.format(branch_name, self.repo_url)
        else:
            clone_cmd = 'git clone --depth 10 {0}'.format(self.repo_url)

        # clone to specific dest folder
        if foldername:
            clone_cmd += " " + foldername
        else:
            # default folder
            dir_name = self.repo_name

        logging.info(clone_cmd)
        ret = util.run_command(clone_cmd, cwd=clone_path, timeout=3600)[0]

        if ret != 0:
            logging.error("clone command return none-zero code!")
            raise CodeSyncError("Failed to clone repo!")

        if self.repo_dir is None:
            self.repo_dir = os.path.join(clone_path, dir_name)

        self.__exec('git config remote.origin.fetch "+refs/heads/*:refs/remotes/origin/*"')
        logging.info('---clone repo successfully---')
        return self.repo_dir

    def checkout_branch(self,
                        branch_name,
                        tag_name=None,
                        submodule_deinit=True,
                        submodule_update=True,
                        shallow_fetch=False):
        """Fetch and checkout to specific branch, and make the branch update to date.

        Execute git commands, and try several times if any steps are failed.
            1. git reset --hard
            2. git clean -fxd
            3. git deinit -f .
            4. git fetch -t origin <branch name>
            5. git checkout <branch name>
            6. git submodule update --init

        Arguments:
            branch_name {str} -- branch name or hash

        Keyword Arguments:
            tag_name {str} -- tag name (default: {None})
            submodule_deinit {bool} -- do submodule deinit (default: {True})
            submodule_update {bool} -- do submodule update (default: {True})
            shallow_fetch {bool} -- do shallow fetch (default: {False})

        Raises:
            BranchError -- When branch is not exists
            CodeSyncError -- When any git operation failed.


        Returns:
            bool -- success or not
        """
        logging.info("start to update repo...")

        if shallow_fetch:
            fetch_command = 'git fetch --depth 10 -t --force origin %s'%branch_name
        else:
            fetch_command = 'git fetch -t --force origin %s'%branch_name

        try:
            if submodule_deinit and submodule_update:
                self.__exec('git submodule deinit -f .')

            self.__exec('git reset --hard', retry=1)
            self.__exec('git clean -fxd', retry=1)
        except CodeSyncError:
            if os.name == "nt":
                util.change_folder_security_win(self.repo_dir)
            Repo.remove_index_lock(self.repo_dir)

            self.__exec('git reset --hard')
            self.__exec('git clean -fxd')

        try:
            self.__exec(fetch_command, retry=4)
        except CodeSyncError:
            raise BranchError("Fetch Branch Error!")

        self.__exec('git checkout -f %s'%branch_name)
        self.__exec('git merge origin/%s'%branch_name)
        if tag_name:
            self.__exec('git checkout %s'%tag_name)

        if submodule_update:
            return self.update_submodules()

    def autosync(self, repo_parent_dir, branchs="master", submodule_deinit=True, submodule_update=True):
        """Automatic sync this repo. Retrun True if sucess.

        This function can be used in bellow situations:
            1. We don't know it is ready in local.
            2. We hope to simple the repo update.

        Arguments:
            repo_parent_dir {str} -- the root directory of the repo is located

        Keyword Arguments:
            branch {str} -- branch name (default: {"master"})
            submodule_deinit {bool} -- do submodule deinit (default: {True})
            submodule_update {bool} -- do submodule update (default: {True})

        Returns:
            bool.
        """
        #Parser branch name and tag name
        branch_name, tag_name = branchs, None
        if "%" in branchs:
            try:
                name_list = branchs.split('%')
                branch_name = name_list[0]
                tag_name = name_list[1]
            except:
                pass

        logging.info("Repo url: %s, Branch: %s, Tag: %s.", self.repo_url, branch_name, tag_name)

        if not os.path.exists(repo_parent_dir):
            os.makedirs(repo_parent_dir)

        # Get repo dir from main repo parent
        if not self.repo_dir:
            self.repo_dir = get_repo_location(repo_parent_dir, self.repo_name, self.repo_url)


        # Clone repo to repo_parent_dir
        if not self.is_ready:
            self.repo_dir = self.clone(repo_parent_dir, branch_name)

        # checkout & update repo
        self.checkout_branch(branch_name, tag_name, submodule_deinit, submodule_update)

        logging.info("---- Sync successfully %s ----", self.repo_dir)
        return True


    @staticmethod
    def remove_index_lock(repo_dir):
        """Remove index.lock to recovery git operation."""

        gitdir = repo_dir+"/.git"
        if os.path.exists(gitdir) is False:
            return

        try:
            for parent, _, filenames in os.walk(gitdir):
                for filename in filenames:
                    if ".lock" in filename:
                        path = os.path.join(parent, filename)
                        os.remove(path)

        except Exception:
            logging.exception("remove git lock failure")


def _repo_is_match(repo_dir, repo_url):
    '''
    check if the specified repo is adentical with the repo_url
    '''
    if os.path.isdir(repo_dir) is False:
        return False
    output = subprocess.check_output('git remote -v', cwd=repo_dir, shell=True)
    return repo_url in output


def get_repo_location(maindir, repo_name, repo_url):
    '''
    Search local repo in maindir,
    Return True: This repo is ready to use.
    Return False: Need Clone.
    '''
    # Search local repo in repo_parent_dir
    dirlist = glob.glob(maindir+"/"+repo_name)
    try:
        path = dirlist[0]
        output = subprocess.check_output('git remote -v', cwd=path, shell=True)
        logging.debug(output)
        return path

    except IndexError:
        logging.warning("Not found repo")
        return None

    except subprocess.CalledProcessError as e:
        logging.error("Repo is damaged. %s", e)
        return None



import os
import os.path as Path
import logging
import subprocess

from elftools.elf.elffile import ELFFile
from elftools.common.exceptions import ELFError

from mcutk.exceptions import ReadElfError
from mcutk.appbase import APPBase
from mcutk.apps import appfactory



def transform_elf_basic(type, in_file, out_file, executor=None):
    """Transform ELF to specific type by arm-none-eabi-objcopy.

    Supported types: bin, ihex, srec.

    Arguments:
        type {str} -- which type you want to convert.
        in_file {str} -- path to elf file.
        out_file {str} -- output file
        executor {str} -- path to arm-none-eabi-objcopy, if it is None,
                    program will use the default executor(mcutk/bin/)

    Raises:
        ReadElfError -- Unknown elf format will raise such error
        Exception -- Convert failed will raise exception

    Returns:
        bool
    """
    supported = {
        'bin': "binary",
        'ihex': 'ihex',
        'srec': 'srec'
    }

    if type not in supported:
        raise ValueError("unknown type, valid choices are: %s"%(str(supported)))

    if executor is None:
        executor = Path.join(Path.dirname(__file__), 'bin/arm-none-eabi-objcopy.exe')

    try:
        with open(in_file, 'rb') as fileobj:
            elffile = ELFFile(fileobj)
            for header_name in elffile.header:
                print('{:<10s} -- {:<10s}'.format(header_name, str(elffile.header[header_name])))
    except ELFError:
        raise ReadElfError('read elf error!')

    if not os.path.isfile(executor):
        raise IOError('arm-none-eabi-objcopy does not exists!')

    type = supported.get(type, type)
    cmds = '{0} -O {1} {2} {3}'.format(executor, type, in_file, out_file)

    return subprocess.call(cmds, shell=True) == 0




def transform_elf(ide, type, in_file, out_file):
    """Transform ELF to specific type with toolchain method (app.transform_elf).

    Supported types: bin, ihex, srec.

    Arguments:
        ide {str} -- ide name or ide app instance.
        type {str} -- which type you want to convert.
        in_file {str} -- path to elf file.
        out_file {str} -- output file

    Raises:
        ReadElfError -- Unknown elf format will raise such error
        Exception -- Convert failed will raise exception

    Returns:
        bool -- success or not.
    """
    if isinstance(ide, APPBase):
        app_instance = ide
    else:
        app_instance = appfactory(ide).get_latest()

    if not (app_instance and app_instance.is_ready):
        raise ValueError('Fatal error, not found toolchain "%s" in system!'%ide)

    return app_instance.transform_elf(type, in_file, out_file)


class ElfFile(object):

    def __init__(self, elffile):
        self.elf = elffile
        self._fileobject = open(self.elf, 'rb')
        self.elffile = ELFFile(self._fileobject)

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self._fileobject.close()

    def get_symbol(self, name):
        """Find and return Symbol object."""
        symbol = None
        name = name.lower()
        for section in self.elffile.iter_sections():
            try:
                for sym in section.iter_symbols():
                    if sym.name.lower() == name:
                        symbol = sym
                        break
            except:
                pass
        return symbol

    def get_section(self, name):
        """Find and return Section object."""
        section = None
        name = name.lower()
        for sec in self.elffile.iter_sections():
            if sec.name.lower() == name:
                section = sec
                break
        return section

    @property
    def sp(self):
        """Parse and return stack pointer from ELF file."""
        search_words = [
            ("symbol", "__VECTOR_TABLE"),
            ("symbol", "_vector_start"),
            ("symbol", "_vector_table"),
            ("symbol", "__vectors_start__"),
            ("symbol", "__vectors"),
            ("section", "VECTOR_ROM"),
        ]
        for sw_type, sw in search_words:
            if sw_type == "symbol":
                s = self.get_symbol(sw)
                if s and s.entry:
                    return s.entry["st_value"]
            else:
                s = self.get_section(sw)
                if s and s.header:
                    return s.header["sh_addr"]

        raise ValueError("not found Vector symbol or section!")

    def get_segments(self):
        """Get segments for load.
        """
        segments = list()
        for segment in self.elffile.iter_segments():
            if segment.header.p_type == 'PT_LOAD' and segment.header.p_filesz != 0:
                logging.info("Writing segment Addr:0x%08x, Vaddr:0x%08x, size %d",
                    segment['p_paddr'], segment['p_vaddr'], segment.header.p_filesz)
                segments.append(segment)
        return segments

    def is_to_flash(self, flash_addr):
        """Retrun a boolean to identify ELF is put to flash.

        Arguments:
            flash_addr: flash start address

        Return:
            Boolean
        """
        # convert to int
        if isinstance(flash_addr, str):
            flash_addr = int(flash_addr, 16)

        to_flash = False
        segments = self.get_segments()
        # check segment addr if is hit the flash area
        for segment in segments:
            if segment['p_paddr'] >= flash_addr:
                to_flash = True
                break
        return to_flash



from __future__ import print_function
import os
import logging
import time
from xml.etree import cElementTree as ET
from collections import defaultdict

import click
from globster import Globster

from mcutk.apps import appfactory
from mcutk.exceptions import ProjectNotFound, ProjectParserError


TOOLCHAINS = [
    'iar',
    'mdk',
    'mcux',
    'armgcc',
    'xcc',
    'codewarrior',
    'lpcx',
]


EXCLUDE_DIR_NAME = [
    'log/',
    'debug/',
    'obj/',
    'release/',
    '.debug/',
    '.release/',
    'RTE/',
    'settings/'
    '.git/',
    '__pycache__/',
    'flexspi_nor_debug/',
    'flexspi_nor_release/'
]

IDE_INS = [appfactory(toolname) for toolname in TOOLCHAINS]

def identify_project(path, toolname=None):
    """ Identify and return initiliazed project instance.

    Arguments:
        path {str} -- project path

    Keyword Arguments:
        toolname {str} -- if toolname is given, it will try to load
            the project with the tool. (default: None)

    Returns:
        Project object
    """
    if toolname:
        cls = appfactory(toolname)
        return cls.Project.frompath(path)

    prj = None
    for cls in IDE_INS:
        try:
            prj = cls.Project.frompath(path)
            break
        except ProjectParserError as e:
            logging.warning(str(e))
        except ET.ParseError as e:
            logging.error("Bad project: %s, Reason: %s", path, e)
        except ProjectNotFound:
            pass

    return prj


def find_projects(root_dir, recursive=True, include_tools=None, exclude_tools=None):
    """Find projects in specific directory.

    Arguments:
        root_dir {string} -- root directory
        recursive {bool} -- recursive mode
        include_tools {list} -- only include specifices tools
        exclude_tools {list} -- exlucde specifices tools
    Returns:
        {dict} -- key: toolchain name, value: a list of Project objects.

    Example:
        >> ps = find_projects("C:/code/mcu-sdk-2.0", True)
        >> ps
        {
            'iar': [<Project Object at 0x1123>, <Project Object at 0x1124>],
            'mdk': [<Project Object at 0x1123>, <Project Object at 0x1124>],
            ...
        }

    """

    projects = defaultdict(list)
    g = Globster(EXCLUDE_DIR_NAME)
    print('Process scanning')
    s_time = time.time()

    project = identify_project(root_dir)
    if project:
        projects[project.idename].append(project)

    if recursive:
        for root, folders, _ in os.walk(root_dir):
            for folder in folders:
                if g.match(folder):
                    continue

                path = os.path.join(root, folder)
                project = identify_project(path)
                if (project and project.idename != "mcux") or (hasattr(project, "is_enabled") and (project.is_enabled)):
                    projects[project.idename].append(project)

    if projects:
        if include_tools:
            projects = {k: v for k, v in projects.items()
                        if k in include_tools}

        elif exclude_tools:
            for toolname in exclude_tools:
                if toolname in projects:
                    projects.pop(toolname)

    e_time = time.time()

    count = 0
    for toolname, prjs in projects.items():
        length = len(prjs)
        count += length

    click.echo("Found projects total {0}, cover {1} toolchains. Used {2:.2f}(s)".format(
        count, len(projects), e_time-s_time))

    for toolname, prjs in projects.items():
        length = len(prjs)
        click.echo(" + {0:<10}  {1}".format(toolname, length))

    return projects, count



import sys
import logging
import click
from mcutk import __version__


class ComplexCLI(click.MultiCommand):

    COMMANDS = [
        'build',
        'scan',
        'config',
        # 'gdbserver',
        'flash'
    ]

    def list_commands(self, ctx):
        return self.COMMANDS

    def get_command(self, ctx, name):
        if sys.version_info[0] == 2:
            name = name.encode('ascii', 'replace')
        mod = __import__('mcutk.commands.' + name, None, None, ['cli'])
        return mod.cli



@click.command(cls=ComplexCLI, invoke_without_command=True, help="mcutk command line tool")
@click.option('-v', '--verbose', is_flag=True, help='show more console message')
@click.option('--version', is_flag=True, help="show mcutk version")
def main(version=False, verbose=False, debug=False):
    if verbose:
        logging.basicConfig(format="[%(levelname)s] %(message)s", level=logging.DEBUG)
    else:
        logging.basicConfig(format="[%(levelname)s] %(message)s", level=logging.WARNING)

    if version:
        click.echo(__version__)


if __name__ == '__main__':
    main()



import sys
import time
import logging
import tempfile
import pexpect
from pexpect.popen_spawn import PopenSpawn

PY = sys.version_info[0]

try:
    basestring
except NameError:
    basestring = str

try:
    unicode
except NameError:
    unicode = str

class GDBSessionInitFailed(Exception):
    pass


class GDBTimeout(Exception):
    pass


class GDBSession(object):
    """GDB debug session manager. This class will start a gdb process in backend.
    And provide methods allow user interact with gdb or manage the state.

    Example:
        >>> session = GDBSession.start("/path/to/gdb <image.elf> -x <gdb.init>")
        >>> response = session.run_cmd("load)
        >>> response = session.run_cmd("continue", timeout=10)
        >>> response = session.run_cmd("q")
        >>> session.close()
        >>> session.console_output
    """


    @staticmethod
    def start(cmdline):
        """A shortcut to start a gdb session.

        Arguments:
            cmdline {str} -- gdb startup command line.
        """

        session = GDBSession(cmdline)
        session.init()
        return session

    def __init__(self, executable):
        """GDB Session constructor.

        Create a gdb debug session. Pass the gdb executable path (also with arguments) as the
        startup command line.

        Arguments:
            executable {str} -- gdb startup command line or gdb executable.
        """

        self.executable = executable
        self._spawn = None
        self._logfile = None
        self._console = ''
        self.initial_arguments = '-ex "set tcp connect-timeout 100"'
        self.timeout = 100
        self._gdbsep = u"\\(gdb\\) "
        self.gdb_server_proc = None

    def init(self):
        """Start GDB process in backend."""

        logging.info(self.executable)
        self._logfile = tempfile.SpooledTemporaryFile(200, mode='w+')
        self._spawn = PopenSpawn(
            self.executable,
            logfile=sys.stdout,
            encoding='utf8',
            timeout=self.timeout
        )
        self._spawn.logfile_read = self._logfile

        try:
            self._spawn.expect(self._gdbsep)
        except Exception as e:
            raise GDBSessionInitFailed(str(e))

    def run_cmd(self, cmd, timeout=-1):
        """Run gdb command.

        Arguments:
            cmd {str} -- gdb command
            timeout {int} -- max timeout to wait the response,
                -1: use default timeout value, None: block and until match.

        Returns:
            {str} -- gdb response text
        """
        response = ''

        # convert to unicode
        cmd = unicode(cmd, "utf-8")
        if not self.is_alive:
            raise RuntimeError("gdb session is inactive, cannot send command.")

        try:
            self._spawn.sendline(cmd)
            self._spawn.expect(self._gdbsep, timeout=timeout)

        except pexpect.TIMEOUT:
            raise GDBTimeout('CMD: %s, wait for response timeout!'%cmd)

        except pexpect.EOF:
            logging.debug("GDB EOF")

        if isinstance(self._spawn.before, basestring):
            response = self._spawn.before

        if isinstance(self._spawn.after, basestring):
            response += self._spawn.after

        return response

    def run_cmds(self, cmds):
        """Run a list of commands."""

        for cmd in cmds:
            self.run_cmd(cmd)

    @property
    def is_alive(self):
        """GDB process is alive or not"""

        return self._spawn.proc.poll() == None

    @property
    def pid(self):
        """GDB process pid"""

        return self._spawn.proc.pid

    def kill(self):
        """Kill GDB process"""

        return self._spawn.proc.kill()

    @property
    def console_output(self):
        """Return all console output
        You must call this, when session is closed.

        """

        if self.is_alive:
            raise RuntimeError('the console output cannot access when session is alive!')
        return self._console

    def _handle_console_output(self):
        self._logfile.seek(0)
        self._console = self._logfile.read()
        self._logfile.close()

    def close(self):
        """Close session and make sure process has exited."""

        if self.is_alive:
            try:
                # send q command to make sure gdb exit
                self._spawn.logfile = None
                self._spawn.sendline('q')
            except IOError:
                pass

            # wait 2 seconds to terminate the gdb process
            start_time = time.time()
            while self.is_alive:
                if time.time() - start_time > 2:
                    self.kill()
                    logging.warn("Terminate GDB process (%s)", self.pid)
                    break

            # wait for exit
            self._spawn.proc.wait()

        self._handle_console_output()
        logging.info("gdb session is closed!")

    def __enter__(self):
        self.init()
        return self

    def __exit__(self, etype, evalue, tb):
        self.close()



"""
APP base class definition.
"""
import abc
import os


class APPBase(object):
    """
    An abstract class representing the interface for an app.
    """
    __metaclass__ = abc.ABCMeta

    # @staticmethod
    # @abc.abstractmethod
    # def get_latest():
    #     pass

    def __init__(self, name, path="", version=None, **kwargs):
        """APPBase interface definition.

        Arguments:
            name {string} -- app name
            path {string} -- app path

        Keyword Arguments:
            version {string} -- app version (default: {None})
        """
        self._name = name
        self._path = path
        self.version = version

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if value in (None, "None", ""):
            raise ValueError("invalid name")
        self._name = value

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, value):
        if value in (None, "None", ""):
            raise ValueError("invalid path")
        self._path = value

    @abc.abstractproperty
    def is_ready(self):
        return os.path.exists(self._path)

    def show(self):
        attrs = vars(self)
        for attr, value in attrs.items():
            print("{0}: {1}".format(attr, value))

    def __str__(self):
        return "App({}-{})".format(self._name, self.version)



import os
from setuptools import setup

version = '0.2.1'

install_requires = [
    'pyocd>=0.14,!=0.24',
    'mbed_ls>=1.7.10',
    'pyserial>=3.4',
    'xmodem',
    'future',
    "pexpect",
    "click>=7.0",
    "pyelftools",
    "pyyaml",
    'globster',
    'enum34'
]

version_file = os.path.join(os.path.dirname(
    os.path.abspath(__file__)), 'mcutk/_version.py')

try:
    with open(version_file, 'w') as f:
        f.write("VERSION='%s'" % version)
except Exception as e:
    print(e)

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="pymcutk",
    version=version,
    url='https://github.com/Hoohaha/pymcutk',
    description="MCU toolkit for mcu automated test.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Haley Guo, Fly Yu",
    license="MIT License",
    author_email="hui.guo@nxp.com",
    install_requires=install_requires,
    packages=["mcutk"],
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'mtk = mcutk.__main__:main',
        ]
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)
